@@ -1,6 +1,6 @@
 /******************************************************************************
  *
- * Copyright (C) 1997-2014 by Dimitri van Heesch.
+ * Copyright (C) 1997-2015 by Dimitri van Heesch.
  *
  * Permission to use, copy, modify, and distribute this software and its
  * documentation under the terms of the GNU General Public License is hereby
@@ -16,19 +16,21 @@
 #ifndef TEMPLATE_H
 #define TEMPLATE_H
 
-#include <qcstring.h>
-#include <qvaluelist.h>
+#include <vector>
 
-class FTextStream;
+#include "qcstring.h"
+#include "containers.h"
+#include "variant.h"
 
 class TemplateListIntf;
 class TemplateStructIntf;
 class TemplateEngine;
+class TextStream;
 
 /** @defgroup template_api Template API
  *
  *  This is the API for a
- *  <a href="https://docs.djangoproject.com/en/1.6/topics/templates/">Django</a>
+ *  <a href="https://www.djangoproject.com/">Django</a>
  *  compatible template system written in C++.
  *  It is somewhat inspired by Stephen Kelly's
  *  <a href="http://www.gitorious.org/grantlee/pages/Home">Grantlee</a>.
@@ -91,34 +93,34 @@ class TemplateVariant
 {
   public:
     /** @brief Helper class to create a delegate that can store a function/method call. */
-    class Delegate
+    class FunctionDelegate
     {
       public:
         /** Callback type to use when creating a delegate from a function. */
-        typedef TemplateVariant (*StubType)(const void *obj, const QValueList<TemplateVariant> &args);
+        typedef TemplateVariant (*StubType)(const void *obj, const std::vector<TemplateVariant> &args);
 
-        Delegate() : m_objectPtr(0) , m_stubPtr(0) {}
+        FunctionDelegate() : m_objectPtr(0) , m_stubPtr(0) {}
 
         /** Creates a delegate given an object. The method to call is passed as a template parameter */
-        template <class T, TemplateVariant (T::*TMethod)(const QValueList<TemplateVariant> &) const>
-        static Delegate fromMethod(const T* objectPtr)
+        template <class T, TemplateVariant (T::*TMethod)(const std::vector<TemplateVariant> &) const>
+        static FunctionDelegate fromMethod(const T* objectPtr)
         {
-          Delegate d;
+          FunctionDelegate d;
           d.m_objectPtr = objectPtr;
           d.m_stubPtr   = &methodStub<T, TMethod>;
           return d;
         }
         /** Creates a delegate given an object, and a plain function. */
-        static Delegate fromFunction(const void *obj,StubType func)
+        static FunctionDelegate fromFunction(const void *obj,StubType func)
         {
-          Delegate d;
+          FunctionDelegate d;
           d.m_objectPtr = obj;
           d.m_stubPtr = func;
           return d;
         }
 
         /** Invokes the function/method stored in the delegate */
-        TemplateVariant operator()(const QValueList<TemplateVariant> &args) const
+        TemplateVariant operator()(const std::vector<TemplateVariant> &args) const
         {
           return (*m_stubPtr)(m_objectPtr, args);
         }
@@ -127,40 +129,31 @@ class TemplateVariant
         const void* m_objectPtr;
         StubType    m_stubPtr;
 
-        template <class T, TemplateVariant (T::*TMethod)(const QValueList<TemplateVariant> &) const>
-        static TemplateVariant methodStub(const void* objectPtr, const QValueList<TemplateVariant> &args)
+        template <class T, TemplateVariant (T::*TMethod)(const std::vector<TemplateVariant> &) const>
+        static TemplateVariant methodStub(const void* objectPtr, const std::vector<TemplateVariant> &args)
         {
           T* p = (T*)(objectPtr);
           return (p->*TMethod)(args);
         }
     };
 
-    /** Types of data that can be stored in a TemplateVariant */
-    enum Type { None, Bool, Integer, String, Struct, List, Function };
-
-    /** Returns the type of the value stored in the variant */
-    Type type() const;
-
-    /** Return a string representation of the type of the value stored in the variant */
-    QCString typeAsString() const;
-
-    /** Returns TRUE if the variant holds a valid value, or FALSE otherwise */
-    bool isValid() const;
-
     /** Constructs an invalid variant. */
-    TemplateVariant();
+    TemplateVariant() {}
 
     /** Constructs a new variant with a boolean value \a b. */
-    explicit TemplateVariant(bool b);
+    explicit TemplateVariant(bool b) { m_variant.set<bool>(b); }
 
     /** Constructs a new variant with a integer value \a v. */
-    TemplateVariant(int v);
+    TemplateVariant(int v) { m_variant.set<int>(v); }
 
     /** Constructs a new variant with a string value \a s. */
-    TemplateVariant(const char *s,bool raw=FALSE);
+    TemplateVariant(const char *s,bool raw=FALSE) : m_raw(raw) { m_variant.set<QCString>(s); }
 
     /** Constructs a new variant with a string value \a s. */
-    TemplateVariant(const QCString &s,bool raw=FALSE);
+    TemplateVariant(const QCString &s,bool raw=FALSE) : m_raw(raw) { m_variant.set<QCString>(s); }
+
+    /** Constructs a new variant with a string value \a s. */
+    TemplateVariant(const std::string &s,bool raw=FALSE) : m_raw(raw) { m_variant.set<QCString>(s); }
 
     /** Constructs a new variant with a struct value \a s.
      *  @note. The variant will hold a reference to the object.
@@ -173,13 +166,13 @@ class TemplateVariant
     TemplateVariant(TemplateListIntf *l);
 
     /** Constructs a new variant which represents a method call
-     *  @param[in] delegate Delegate object to invoke when
+     *  @param[in] delegate FunctionDelegate object to invoke when
      *             calling call() on this variant.
-     *  @note Use TemplateVariant::Delegate::fromMethod() and
-     *  TemplateVariant::Delegate::fromFunction() to create
-     *  Delegate objects.
+     *  @note Use TemplateVariant::FunctionDelegate::fromMethod() and
+     *  TemplateVariant::FunctionDelegate::fromFunction() to create
+     *  FunctionDelegate objects.
      */
-    TemplateVariant(const Delegate &delegate);
+    TemplateVariant(const FunctionDelegate &delegate) { m_variant.set<FunctionDelegate>(delegate); }
 
     /** Destroys the Variant object */
     ~TemplateVariant();
@@ -189,52 +182,157 @@ class TemplateVariant
      */
     TemplateVariant(const TemplateVariant &v);
 
+    /** Moves the contents of variant \a v into this variant.
+     *  variant \a v will become invalid
+     */
+    TemplateVariant(TemplateVariant &&v);
+
     /** Assigns the value of the variant \a v to this variant. */
     TemplateVariant &operator=(const TemplateVariant &v);
 
+    /** Move the value of the variant \a v into this variant.
+     *  Variant \a v will become invalid */
+    TemplateVariant &operator=(TemplateVariant &&v);
+
     /** Compares this QVariant with v and returns true if they are equal;
      *  otherwise returns false.
      */
-    bool operator==(TemplateVariant &other);
+    bool operator==(TemplateVariant &other)
+    {
+      if (!m_variant.valid())
+      {
+        return FALSE;
+      }
+      if (isBool() && other.isBool())
+      {
+        return m_variant.get<bool>() == other.m_variant.get<bool>();
+      }
+      else if (isInt() && other.isInt())
+      {
+        return m_variant.get<int>() == other.m_variant.get<int>();
+      }
+      else if (isList() && other.isList())
+      {
+        return m_variant.get<TemplateListIntf*>() == other.m_variant.get<TemplateListIntf*>();
+      }
+      else if (isStruct() && other.isStruct())
+      {
+        return m_variant.get<TemplateStructIntf*>() == other.m_variant.get<TemplateStructIntf*>();
+      }
+      return toString()==other.toString();
+    }
+
+    QCString listToString() const;
+    QCString structToString() const;
 
     /** Returns the variant as a string. */
-    QCString            toString() const;
+    QCString toString() const;
 
     /** Returns the variant as a boolean. */
-    bool                toBool() const;
+    bool toBool() const;
 
     /** Returns the variant as an integer. */
-    int                 toInt() const;
+    int toInt() const;
+
+    /** Returns TRUE if the variant holds a valid value, or FALSE otherwise */
+    constexpr bool isValid()    const { return m_variant.valid(); }
+    /** Returns TRUE if the variant holds a boolean value */
+    constexpr bool isBool()     const { return m_variant.is<bool>(); }
+    /** Returns TRUE if the variant holds an integer value */
+    constexpr bool isInt()      const { return m_variant.is<int>(); }
+    /** Returns TRUE if the variant holds a string value */
+    constexpr bool isString()   const { return m_variant.is<QCString>(); }
+    /** Returns TRUE if the variant holds a struct value */
+    constexpr bool isStruct()   const { return m_variant.is<TemplateStructIntf*>(); }
+    /** Returns TRUE if the variant holds a list value */
+    constexpr bool isList()     const { return m_variant.is<TemplateListIntf*>(); }
+    /** Returns TRUE if the variant holds a function value */
+    constexpr bool isFunction() const { return m_variant.is<FunctionDelegate>(); }
 
     /** Returns the pointer to list referenced by this variant
      *  or 0 if this variant does not have list type.
      */
-    TemplateListIntf   *toList() const;
+    TemplateListIntf   *toList()
+    {
+      return isList() ? m_variant.get<TemplateListIntf*>() : nullptr;
+    }
+    const TemplateListIntf *toList() const
+    {
+      return isList() ? m_variant.get<TemplateListIntf*>() : nullptr;
+    }
 
     /** Returns the pointer to struct referenced by this variant
      *  or 0 if this variant does not have struct type.
      */
-    TemplateStructIntf *toStruct() const;
+    TemplateStructIntf *toStruct()
+    {
+      return isStruct() ? m_variant.get<TemplateStructIntf*>() : nullptr;
+    }
+    const TemplateStructIntf *toStruct() const
+    {
+      return isStruct() ? m_variant.get<TemplateStructIntf*>() : nullptr;
+    }
 
     /** Return the result of apply this function with \a args.
      *  Returns an empty string if the variant type is not a function.
      */
-    TemplateVariant call(const QValueList<TemplateVariant> &args);
+    TemplateVariant call(const std::vector<TemplateVariant> &args)
+    {
+      return isFunction() ? m_variant.get<FunctionDelegate>()(args) : TemplateVariant();
+    }
 
     /** Sets whether or not the value of the Variant should be
      *  escaped or written as-is (raw).
      *  @param[in] b TRUE means write as-is, FALSE means apply escaping.
      */
-    void setRaw(bool b);
+    void setRaw(bool b) { m_raw = b; }
 
     /** Returns whether or not the value of the Value is raw.
      *  @see setRaw()
      */
-    bool raw() const;
+    constexpr bool raw() const { return m_raw; }
+
+    /** Symbolic names for the possible types that this variant can hold. */
+    enum class Type : size_t
+    {
+      None     = std::string::npos,
+      Bool     = 0,
+      Int      = 1,
+      String   = 2,
+      Struct   = 3,
+      List     = 4,
+      Function = 5
+    };
+
+    /** Returns the type held by this variant */
+    constexpr Type type() const { return static_cast<Type>(m_variant.index()); }
+
+    /** Return a string representation of the type of the value stored in the variant */
+    constexpr const char *typeAsString() const
+    {
+      switch (type())
+      {
+        case Type::None:       return "invalid";
+        case Type::Bool:       return "bool";
+        case Type::Int:        return "integer";
+        case Type::String:     return "string";
+        case Type::Struct:     return "struct";
+        case Type::List:       return "list";
+        case Type::Function:   return "function";
+      }
+      return "invalid";
+    }
 
   private:
-    class Private;
-    Private *p;
+    using VariantT = Variant<bool,                  // index==0: Type::Bool
+                             int,                   // index==1: Type::Int
+                             QCString,              // index==2: Type::String
+                             TemplateStructIntf*,   // index==3: Type::Struct
+                             TemplateListIntf*,     // index==4: Type::List
+                             FunctionDelegate       // index==5: Type::Function
+                            >;
+    VariantT              m_variant;
+    bool                  m_raw = false;
 };
 
 //------------------------------------------------------------------------
@@ -292,10 +390,10 @@ class TemplateListIntf
     virtual ~TemplateListIntf() {}
 
     /** Returns the number of elements in the list */
-    virtual int count() const = 0;
+    virtual uint count() const = 0;
 
     /** Returns the element at index position \a index. */
-    virtual TemplateVariant  at(int index) const = 0;
+    virtual TemplateVariant  at(uint index) const = 0;
 
     /** Creates a new iterator for this list.
      *  @note the user should call delete on the returned pointer.
@@ -305,7 +403,7 @@ class TemplateListIntf
     /** Increase object's reference count */
     virtual int addRef() = 0;
 
-    /** Decreases object's referenc count, destroy object if 0 */
+    /** Decreases object's reference count, destroy object if 0 */
     virtual int release() = 0;
 };
 
@@ -314,8 +412,8 @@ class TemplateList : public TemplateListIntf
 {
   public:
     // TemplateListIntf methods
-    virtual int  count() const;
-    virtual TemplateVariant at(int index) const;
+    virtual uint count() const;
+    virtual TemplateVariant at(uint index) const;
     virtual TemplateListIntf::ConstIterator *createIterator() const;
     virtual int addRef();
     virtual int release();
@@ -349,12 +447,15 @@ class TemplateStructIntf
     /** Gets the value for a field name.
      *  @param[in] name The name of the field.
      */
-    virtual TemplateVariant get(const char *name) const = 0;
+    virtual TemplateVariant get(const QCString &name) const = 0;
+
+    /** Return the list of fields. */
+    virtual StringVector fields() const = 0;
 
     /** Increase object's reference count */
     virtual int addRef() = 0;
 
-    /** Decreases object's referenc count, destroy object if 0 */
+    /** Decreases object's reference count, destroy object if 0 */
     virtual int release() = 0;
 };
 
@@ -364,7 +465,8 @@ class TemplateStruct : public TemplateStructIntf
 {
   public:
     // TemplateStructIntf methods
-    virtual TemplateVariant get(const char *name) const;
+    virtual TemplateVariant get(const QCString &name) const;
+    virtual StringVector fields() const;
     virtual int addRef();
     virtual int release();
 
@@ -375,7 +477,7 @@ class TemplateStruct : public TemplateStructIntf
      *  @param[in] name The name of the field.
      *  @param[in] v The value to set.
      */
-    virtual void set(const char *name,const TemplateVariant &v);
+    virtual void set(const QCString &name,const TemplateVariant &v);
 
 
   private:
@@ -394,8 +496,11 @@ class TemplateStruct : public TemplateStructIntf
 class TemplateEscapeIntf
 {
   public:
+    virtual ~TemplateEscapeIntf() {}
     /** Returns the \a input after escaping certain characters */
     virtual QCString escape(const QCString &input) = 0;
+    /** Setting tabbing mode on or off (for LaTeX) */
+    virtual void enableTabbing(bool b) = 0;
 };
 
 //------------------------------------------------------------------------
@@ -404,6 +509,7 @@ class TemplateEscapeIntf
 class TemplateSpacelessIntf
 {
   public:
+    virtual ~TemplateSpacelessIntf() {}
     /** Returns the \a input after removing redundant whitespace */
     virtual QCString remove(const QCString &input) = 0;
     /** Reset filter state */
@@ -438,7 +544,7 @@ class TemplateContext
      *  @note When a given key is already present,
      *  its value will be replaced by \a v
      */
-    virtual void set(const char *name,const TemplateVariant &v) = 0;
+    virtual void set(const QCString &name,const TemplateVariant &v) = 0;
 
     /** Gets the value for a given key
      *  @param[in] name The name of key.
@@ -453,7 +559,7 @@ class TemplateContext
      */
     virtual const TemplateVariant *getRef(const QCString &name) const = 0;
 
-    /** When files are create (i.e. by {% create ... %}) they written
+    /** When files are created (i.e. by {% create ... %}) they written
      *  to the directory \a dir.
      */
     virtual void setOutputDirectory(const QCString &dir) = 0;
@@ -485,7 +591,7 @@ class Template
      *  @param[in] c The context containing data that can be used
      *  when instantiating the template.
      */
-    virtual void render(FTextStream &ts,TemplateContext *c) = 0;
+    virtual void render(TextStream &ts,TemplateContext *c) = 0;
 };
 
 //------------------------------------------------------------------------
@@ -510,11 +616,10 @@ class TemplateEngine
      */
     void destroyContext(TemplateContext *ctx);
 
-    /** Creates a new template whole contents are in a file.
-     *  @param[in] fileName The name of the file containing the
-     *             template data
+    /** Creates a new template whose contents are in a file.
+     *  @param[in] fileName The name of the file containing the template data
      *  @param[in] fromLine The line number of the statement that triggered the load
-     *  @return the new template, the caller will be the owner.
+     *  @return the new template, the engine will keep ownership of the object.
      */
     Template *loadByName(const QCString &fileName,int fromLine);
 
@@ -523,13 +628,27 @@ class TemplateEngine
      */
     void unload(Template *t);
 
-    void printIncludeContext(const char *fileName,int line) const;
+    /** Prints the current template file include stack */
+    void printIncludeContext(const QCString &fileName,int line) const;
+
+    /** Sets the search directory where to look for template files */
+    void setTemplateDir(const QCString &dirName);
 
   private:
     friend class TemplateNodeBlock;
+    friend class TemplateNodeCreate;
+
     void enterBlock(const QCString &fileName,const QCString &blockName,int line);
     void leaveBlock();
 
+    /** Sets the extension of the output file. This is used to control the
+     *  format of 'special' tags in the template
+     */
+    void setOutputExtension(const QCString &extension);
+
+    /** Returns the output extension, set via setOutputExtension() */
+    QCString outputExtension() const;
+
     class Private;
     Private *p;
 };
